from typing import Any, Union, List, Generic, TypeVar, cast

import entrypoints
from toolz import curry


PluginType = TypeVar('PluginType')


class PluginEnabler(object):
    """Context manager for enabling plugins

    This object lets you use enable() as a context manager to
    temporarily enable a given plugin::

        with plugins.enable('name'):
            do_something()  # 'name' plugin temporarily enabled
        # plugins back to original state
    """
    def __init__(self, registry, name, **options):
        self.registry = registry
        self.name = name
        self.options = options
        self.original_state = registry._get_state()
        self.registry._enable(name, **options)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.registry._set_state(self.original_state)

    def __repr__(self):
        return "{}.enable({!r})".format(self.registry.__class__.__name__, self.name)


class PluginRegistry(Generic[PluginType]):
    """A registry for plugins.

    This is a plugin registry that allows plugins to be loaded/registered
    in two ways:

    1. Through an explicit call to ``.register(name, value)``.
    2. By looking for other Python packages that are installed and provide
       a setuptools entry point group.

    When you create an instance of this class, provide the name of the
    entry point group to use::

        reg = PluginRegister('my_entrypoint_group')

    """
    # this is a mapping of name to error message to allow custom error messages
    # in case an entrypoint is not found
    entrypoint_err_messages = {}

    # global settings is a key-value mapping of settings that are stored globally
    # in the registry rather than passed to the plugins
    _global_settings = {}

    def __init__(self, entry_point_group='', plugin_type=object):
        # type: (str, Any) -> None
        """Create a PluginRegistry for a named entry point group.

        Parameters
        ==========
        entry_point_group: str
            The name of the entry point group.
        plugin_type: object
            A type that will optionally be used for runtime type checking of
            loaded plugins using isinstance.
        """
        self.entry_point_group = entry_point_group
        self.plugin_type = plugin_type
        self._active = None     # type: str
        self._active_name = ''  # type: str
        self._plugins = {}      # type: dict
        self._options = {}      # type: dict
        self._global_settings = self.__class__._global_settings.copy()  # type: dict

    def register(self, name, value):
        # type: (str, Union[PluginType, None]) -> PluginType
        """Register a plugin by name and value.

        This method is used for explicit registration of a plugin and shouldn't be
        used to manage entry point managed plugins, which are auto-loaded.

        Parameters
        ==========
        name: str
            The name of the plugin.
        value: PluginType or None
            The actual plugin object to register or None to unregister that plugin.

        Returns
        =======
        plugin: PluginType
            The plugin that was registered or unregistered.
        """
        if value is None and name in self._plugins:
            return self._plugins.pop(name)
        else:
            assert isinstance(value, self.plugin_type)
            self._plugins[name] = value
            return value

    def names(self):
        # type: () -> List[str]
        """List the names of the registered and entry points plugins."""
        exts = list(self._plugins.keys())
        more_exts = [ep.name for ep in entrypoints.get_group_all(self.entry_point_group)]
        exts.extend(more_exts)
        return sorted(set(exts))

    def _get_state(self):
        """Return a dictionary representing the current state of the registry"""
        return {'_active': self._active,
                '_active_name': self._active_name,
                '_plugins': self._plugins.copy(),
                '_options': self._options.copy(),
                '_global_settings': self._global_settings.copy()}

    def _set_state(self, state):
        """Reset the state of the registry"""
        assert set(state.keys()) == {'_active', '_active_name',
                                     '_plugins', '_options', '_global_settings'}
        for key, val in state.items():
            setattr(self, key, val)

    def _enable(self, name, **options):
        # type: (str, **Any) -> None
        if name not in self._plugins:
            try:
                ep = entrypoints.get_single(self.entry_point_group, name)
            except entrypoints.NoSuchEntryPoint:
                if name in self.entrypoint_err_messages:
                    raise ValueError(self.entrypoint_err_messages[name])
                else:
                    raise
            value = cast(PluginType, ep.load())
            assert isinstance(value, self.plugin_type)
            self.register(name, value)
        self._active_name = name
        self._active = self._plugins[name]
        for key in set(options.keys()) & set(self._global_settings.keys()):
            self._global_settings[key] = options.pop(key)
        self._options = options

    def enable(self, name=None, **options):
        # type: (str, **Any) -> PluginEnabler
        """Enable a plugin by name.

        This can be either called directly, or used as a context manager.

        Parameters
        ----------
        name : string (optional)
            The name of the plugin to enable. If not specified, then use the
            current active name.
        **options :
            Any additional parameters will be passed to the plugin as keyword
            arguments

        Returns
        -------
        PluginEnabler:
            An object that allows enable() to be used as a context manager
        """
        if name is None:
            name = self.active
        return PluginEnabler(self, name, **options)


    @property
    def active(self):
        # type: () -> str
        """Return the name of the currently active plugin"""
        return self._active_name

    @property
    def options(self):
        # type: () -> str
        """Return the current options dictionary"""
        return self._options

    def get(self):
        # type: () -> PluginType
        """Return the currently active plugin."""
        if self._options:
            return curry(self._active, **self._options)
        else:
            return self._active

    def __repr__(self):
        # type: () -> str
        return ("{}(active={!r}, registered={!r})"
                "".format(self.__class__.__name__,
                          self._active_name,
                          list(self.names())))



# flake8: noqa
from .display import vega, Vega, renderers
from .schema import *

from .data import (
    pipe, curry, limit_rows,
    sample, to_json, to_csv, to_values,
    default_data_transformer
)



"""
Bar Chart with Highlighted Bar
------------------------------
This example shows a basic bar chart with a single bar highlighted.
"""
# category: bar charts
import altair as alt
from vega_datasets import data

source = data.wheat()

alt.Chart(source).mark_bar().encode(
    x='year:O',
    y="wheat:Q",
    # The highlight will be set on the result of a conditional statement
    color=alt.condition(
        alt.datum.year == 1810,  # If the year is 1810 this test returns True,
        alt.value('orange'),     # which sets the bar orange.
        alt.value('steelblue')   # And if it's not true it sets the bar steelblue.
    )
).properties(width=600)



import pytest
import json

try:
    import selenium
except ImportError:
    selenium = None

from ..mimebundle import spec_to_mimebundle


# example from https://vega.github.io/editor/#/examples/vega-lite/bar
VEGALITE_SPEC = json.loads(
    """
        {
    "$schema": "https://vega.github.io/schema/vega-lite/v2.json",
    "description": "A simple bar chart with embedded data.",
    "data": {
        "values": [
        {"a": "A","b": 28}, {"a": "B","b": 55}, {"a": "C","b": 43},
        {"a": "D","b": 91}, {"a": "E","b": 81}, {"a": "F","b": 53},
        {"a": "G","b": 19}, {"a": "H","b": 87}, {"a": "I","b": 52}
        ]
    },
    "mark": "bar",
    "encoding": {
        "x": {"field": "a", "type": "ordinal"},
        "y": {"field": "b", "type": "quantitative"}
    }
    }
    """
)

VEGA_SPEC = json.loads(
    """
    {
    "$schema": "https://vega.github.io/schema/vega/v3.0.json",
    "description": "A simple bar chart with embedded data.",
    "autosize": "pad",
    "padding": 5,
    "height": 200,
    "style": "cell",
    "data": [
        {
        "name": "source_0",
        "values": [
            {"a": "A", "b": 28},
            {"a": "B", "b": 55},
            {"a": "C", "b": 43},
            {"a": "D", "b": 91},
            {"a": "E", "b": 81},
            {"a": "F", "b": 53},
            {"a": "G", "b": 19},
            {"a": "H", "b": 87},
            {"a": "I", "b": 52}
        ]
        },
        {
        "name": "data_0",
        "source": "source_0",
        "transform": [
            {"type": "formula", "expr": "toNumber(datum[\\"b\\"])", "as": "b"},
            {
            "type": "filter",
            "expr": "datum[\\"b\\"] !== null && !isNaN(datum[\\"b\\"])"
            }
        ]
        }
    ],
    "signals": [
        {"name": "x_step", "value": 21},
        {
        "name": "width",
        "update": "bandspace(domain('x').length, 0.1, 0.05) * x_step"
        }
    ],
    "marks": [
        {
        "name": "marks",
        "type": "rect",
        "style": ["bar"],
        "from": {"data": "data_0"},
        "encode": {
            "update": {
            "fill": {"value": "#4c78a8"},
            "x": {"scale": "x", "field": "a"},
            "width": {"scale": "x", "band": true},
            "y": {"scale": "y", "field": "b"},
            "y2": {"scale": "y", "value": 0}
            }
        }
        }
    ],
    "scales": [
        {
        "name": "x",
        "type": "band",
        "domain": {"data": "data_0", "field": "a", "sort": true},
        "range": {"step": {"signal": "x_step"}},
        "paddingInner": 0.1,
        "paddingOuter": 0.05
        },
        {
        "name": "y",
        "type": "linear",
        "domain": {"data": "data_0", "field": "b"},
        "range": [{"signal": "height"}, 0],
        "nice": true,
        "zero": true
        }
    ],
    "axes": [
        {
        "scale": "x",
        "orient": "bottom",
        "title": "a",
        "labelOverlap": true,
        "encode": {
            "labels": {
            "update": {
                "angle": {"value": 270},
                "align": {"value": "right"},
                "baseline": {"value": "middle"}
            }
            }
        },
        "zindex": 1
        },
        {
        "scale": "y",
        "orient": "left",
        "title": "b",
        "labelOverlap": true,
        "tickCount": {"signal": "ceil(height/40)"},
        "zindex": 1
        },
        {
        "scale": "y",
        "orient": "left",
        "grid": true,
        "tickCount": {"signal": "ceil(height/40)"},
        "gridScale": "x",
        "domain": false,
        "labels": false,
        "maxExtent": 0,
        "minExtent": 0,
        "ticks": false,
        "zindex": 0
        }
    ],
    "config": {"axisY": {"minExtent": 30}}
    }
    """
)

VEGAEMBED_VERSION = '3.14.0'
VEGALITE_VERSION = '2.3.1'
VEGA_VERSION = '3.3.1'

@pytest.mark.skipif('not selenium')
def test_spec_to_vega_mimebundle():
    try:
        bundle = spec_to_mimebundle(
            spec=VEGALITE_SPEC,
            format='vega',
            mode='vega-lite',
            vega_version=VEGA_VERSION,
            vegalite_version=VEGALITE_VERSION,
            vegaembed_version=VEGAEMBED_VERSION
        )
    except ValueError as err:
        if str(err).startswith('Internet connection'):
            pytest.skip("web connection required for png/svg export")
        else:
            raise
    assert bundle == {'application/vnd.vega.v3+json': VEGA_SPEC}


def test_spec_to_vegalite_mimebundle():
    bundle = spec_to_mimebundle(
        spec=VEGALITE_SPEC,
        mode='vega-lite',
        format='vega-lite',
        vegalite_version=VEGALITE_VERSION
    )
    assert bundle == {'application/vnd.vegalite.v2+json': VEGALITE_SPEC}


def test_spec_to_json_mimebundle():
    bundle = spec_to_mimebundle(
        spec=VEGALITE_SPEC,
        mode='vega-lite',
        format='json',
    )
    assert bundle == {'application/json': VEGALITE_SPEC}



# -*- coding: utf-8 -*-
#
# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.
from . import core
from altair.utils import use_signature
from altair.utils.schemapi import Undefined


class MarkMethodMixin(object):
    """A mixin class that defines mark methods"""

    def mark_area(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'area'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="area", **kwds)
        else:
            copy.mark = "area"
        return copy

    def mark_bar(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                 clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                 dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                 fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                 fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                 limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                 radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                 strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                 strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                 strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                 theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                 xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'bar'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="bar", **kwds)
        else:
            copy.mark = "bar"
        return copy

    def mark_line(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'line'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="line", **kwds)
        else:
            copy.mark = "line"
        return copy

    def mark_trail(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                   clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                   dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                   fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                   fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                   limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined,
                   point=Undefined, radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                   strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                   strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                   strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                   theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                   xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'trail'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="trail", **kwds)
        else:
            copy.mark = "trail"
        return copy

    def mark_point(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                   clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                   dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                   fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                   fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                   limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined,
                   point=Undefined, radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                   strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                   strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                   strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                   theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                   xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'point'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="point", **kwds)
        else:
            copy.mark = "point"
        return copy

    def mark_text(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'text'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="text", **kwds)
        else:
            copy.mark = "text"
        return copy

    def mark_tick(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'tick'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="tick", **kwds)
        else:
            copy.mark = "tick"
        return copy

    def mark_rect(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'rect'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="rect", **kwds)
        else:
            copy.mark = "rect"
        return copy

    def mark_rule(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                  clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                  dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined, point=Undefined,
                  radius=Undefined, shape=Undefined, size=Undefined, stroke=Undefined,
                  strokeCap=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                  strokeJoin=Undefined, strokeMiterLimit=Undefined, strokeOpacity=Undefined,
                  strokeWidth=Undefined, style=Undefined, tension=Undefined, text=Undefined,
                  theta=Undefined, thickness=Undefined, tooltip=Undefined, x2Offset=Undefined,
                  xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'rule'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="rule", **kwds)
        else:
            copy.mark = "rule"
        return copy

    def mark_circle(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                    clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                    dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                    fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                    fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                    limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined,
                    point=Undefined, radius=Undefined, shape=Undefined, size=Undefined,
                    stroke=Undefined, strokeCap=Undefined, strokeDash=Undefined,
                    strokeDashOffset=Undefined, strokeJoin=Undefined, strokeMiterLimit=Undefined,
                    strokeOpacity=Undefined, strokeWidth=Undefined, style=Undefined, tension=Undefined,
                    text=Undefined, theta=Undefined, thickness=Undefined, tooltip=Undefined,
                    x2Offset=Undefined, xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'circle'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="circle", **kwds)
        else:
            copy.mark = "circle"
        return copy

    def mark_square(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                    clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                    dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                    fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                    fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                    limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined,
                    point=Undefined, radius=Undefined, shape=Undefined, size=Undefined,
                    stroke=Undefined, strokeCap=Undefined, strokeDash=Undefined,
                    strokeDashOffset=Undefined, strokeJoin=Undefined, strokeMiterLimit=Undefined,
                    strokeOpacity=Undefined, strokeWidth=Undefined, style=Undefined, tension=Undefined,
                    text=Undefined, theta=Undefined, thickness=Undefined, tooltip=Undefined,
                    x2Offset=Undefined, xOffset=Undefined, y2Offset=Undefined, yOffset=Undefined, **kwds):
        """Set the chart's mark to 'square'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="square", **kwds)
        else:
            copy.mark = "square"
        return copy

    def mark_geoshape(self, align=Undefined, angle=Undefined, baseline=Undefined, binSpacing=Undefined,
                      clip=Undefined, color=Undefined, cornerRadius=Undefined, cursor=Undefined,
                      dir=Undefined, dx=Undefined, dy=Undefined, ellipsis=Undefined, fill=Undefined,
                      fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                      fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                      limit=Undefined, line=Undefined, opacity=Undefined, orient=Undefined,
                      point=Undefined, radius=Undefined, shape=Undefined, size=Undefined,
                      stroke=Undefined, strokeCap=Undefined, strokeDash=Undefined,
                      strokeDashOffset=Undefined, strokeJoin=Undefined, strokeMiterLimit=Undefined,
                      strokeOpacity=Undefined, strokeWidth=Undefined, style=Undefined,
                      tension=Undefined, text=Undefined, theta=Undefined, thickness=Undefined,
                      tooltip=Undefined, x2Offset=Undefined, xOffset=Undefined, y2Offset=Undefined,
                      yOffset=Undefined, **kwds):
        """Set the chart's mark to 'geoshape'
    
        For information on additional arguments, see :class:`MarkDef`
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, binSpacing=binSpacing, clip=clip,
                    color=color, cornerRadius=cornerRadius, cursor=cursor, dir=dir, dx=dx, dy=dy,
                    ellipsis=ellipsis, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, line=line, opacity=opacity, orient=orient,
                    point=point, radius=radius, shape=shape, size=size, stroke=stroke,
                    strokeCap=strokeCap, strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                    strokeJoin=strokeJoin, strokeMiterLimit=strokeMiterLimit,
                    strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, style=style, tension=tension,
                    text=text, theta=theta, thickness=thickness, tooltip=tooltip, x2Offset=x2Offset,
                    xOffset=xOffset, y2Offset=y2Offset, yOffset=yOffset, **kwds)
        copy = self.copy(deep=False)
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="geoshape", **kwds)
        else:
            copy.mark = "geoshape"
        return copy


class ConfigMethodMixin(object):
    """A mixin class that defines config methods"""

    @use_signature(core.Config)
    def configure(self, *args, **kwargs):
        copy = self.copy(deep=False)
        copy.config = core.Config(*args, **kwargs)
        return copy

    @use_signature(core.AreaConfig)
    def configure_area(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["area"] = core.AreaConfig(*args, **kwargs)
        return copy

    @use_signature(core.AxisConfig)
    def configure_axis(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axis"] = core.AxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisBand(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisBand"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisBottom(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisBottom"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisLeft(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisLeft"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisRight(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisRight"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisTop(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisTop"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisX(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisX"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisY(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["axisY"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.BarConfig)
    def configure_bar(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["bar"] = core.BarConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_circle(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["circle"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_geoshape(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["geoshape"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.HeaderConfig)
    def configure_header(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["header"] = core.HeaderConfig(*args, **kwargs)
        return copy

    @use_signature(core.LegendConfig)
    def configure_legend(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["legend"] = core.LegendConfig(*args, **kwargs)
        return copy

    @use_signature(core.LineConfig)
    def configure_line(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["line"] = core.LineConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_mark(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["mark"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_point(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["point"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.ProjectionConfig)
    def configure_projection(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["projection"] = core.ProjectionConfig(*args, **kwargs)
        return copy

    @use_signature(core.RangeConfig)
    def configure_range(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["range"] = core.RangeConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_rect(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["rect"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_rule(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["rule"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.ScaleConfig)
    def configure_scale(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["scale"] = core.ScaleConfig(*args, **kwargs)
        return copy

    @use_signature(core.SelectionConfig)
    def configure_selection(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["selection"] = core.SelectionConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_square(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["square"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.TextConfig)
    def configure_text(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["text"] = core.TextConfig(*args, **kwargs)
        return copy

    @use_signature(core.TickConfig)
    def configure_tick(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["tick"] = core.TickConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgTitleConfig)
    def configure_title(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["title"] = core.VgTitleConfig(*args, **kwargs)
        return copy

    @use_signature(core.LineConfig)
    def configure_trail(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["trail"] = core.LineConfig(*args, **kwargs)
        return copy

    @use_signature(core.ViewConfig)
    def configure_view(self, *args, **kwargs):
        copy = self.copy(deep=['config'])
        if copy.config is Undefined:
            copy.config = core.Config()
        copy.config["view"] = core.ViewConfig(*args, **kwargs)
        return copy


"""
Dot Dash Plot
=============
How to make the dot-dash plot presented in Edward Tufte's `Visual Display of Quantitative Information <https://www.edwardtufte.com/tufte/books_vdqi>`_. Based
on a JavaScript implementation by `g3o2 <https://bl.ocks.org/g3o2/bd4362574137061c243a2994ba648fb8>`_.
"""
# category: scatter plots
import altair as alt
from vega_datasets import data

source = data.cars()

# Configure the options common to all layers
brush = alt.selection(type='interval')
base = alt.Chart(source).add_selection(brush)

# Configure the points
points = base.mark_point().encode(
    x=alt.X('Miles_per_Gallon', title=''),
    y=alt.Y('Horsepower', title=''),
    color=alt.condition(brush, 'Origin', alt.value('grey'))
)

# Configure the ticks
tick_axis = alt.Axis(labels=False, domain=False, ticks=False)

x_ticks = base.mark_tick().encode(
    alt.X('Miles_per_Gallon', axis=tick_axis),
    alt.Y('Origin', title='', axis=tick_axis),
    color=alt.condition(brush, 'Origin', alt.value('lightgrey'))
)

y_ticks = base.mark_tick().encode(
    alt.X('Origin', title='', axis=tick_axis),
    alt.Y('Horsepower', axis=tick_axis),
    color=alt.condition(brush, 'Origin', alt.value('lightgrey'))
)

# Build the chart
y_ticks | (points & x_ticks)



"""
Bar Chart with Line on Dual Axis
--------------------------------
This example shows how to combine two plots and keep their axes.

For a more polished version of this chart, see :ref:`gallery_wheat_wages`.
"""
# category: bar charts
import altair as alt
from vega_datasets import data

source = data.wheat()

base = alt.Chart(source).encode(x='year:O')

bar = base.mark_bar().encode(y='wheat:Q')

line =  base.mark_line(color='red').encode(
    y='wages:Q'
)

(bar + line).properties(width=600)



"""
Seattle Weather Heatmap
-----------------------
This example shows the 2010 daily high temperature (F) in Seattle, WA.
"""
# category: case studies
import altair as alt
from vega_datasets import data

# Since the data is more than 5,000 rows we'll import it from a URL
source = data.seattle_temps.url

alt.Chart(
    source,
    title="2010 Daily High Temperature (F) in Seattle, WA"
).mark_rect().encode(
    x='date(date):O',
    y='month(date):O',
    color=alt.Color('max(temp):Q', scale=alt.Scale(scheme="inferno")),
    tooltip=[
        alt.Tooltip('monthdate(date):T', title='Date'),
        alt.Tooltip('max(temp):Q', title='Max Temp')
    ]
).properties(width=550)
